-- Lattice theory, from Dijkstra
-- Author:: Thierry Coquand

D : Set
(<=) : D -> D -> Prop
meet : D -> D -> D
sup : D -> D -> D

infix 6 <=

ref  : (x : D) -> (y:D) -> (x <= y /\ y <= x   <=>   Eq D x y)
axM : (x : D) -> (y : D) -> (z : D) ->
	      ( z <= meet x y	<=>
		z <= x /\ z <= y
	      )
axS : (x : D) -> (y : D) -> (z : D) ->
	      ( sup x y	<= z <=>
		x <= z /\ y <= z
	      )

lem1M : (x : D) -> (y : D) -> meet x y <= x
lem1M = \x -> \y -> fol-plugin (ref, axM)

lem1S : (x : D) -> (y : D) -> x <= sup x y
lem1S = \x -> \y -> fol-plugin (ref, axS)


assocM : (x : D) -> (y : D) -> (z:D) -> Eq D (meet (meet x y) z) (meet x (meet y z))
assocM = \x -> \y -> \z -> fol-plugin (ref, axM)

lem2M : (x:D) -> (y:D) -> ( x <= y    <=>  Eq D (meet x y) x )
lem2M =  \x -> \y -> fol-plugin (ref, axM)


trans : (x:D) -> (y:D) -> (z:D) -> x <= y /\ y <= z ==> x <= z
trans = \ x -> \ y -> \ z -> \ h1 ->
        let
          rem1 : Eq D (meet (meet x y) z) (meet x z) 
          rem1 = fol-plugin (lem2M,h1)
          rem2 : Eq D (meet x (meet y z)) x
          rem2 = fol-plugin (lem2M,h1)
          rem3 : Eq D (meet x z) x
          rem3 = fol-plugin (rem1,rem2,assocM)
        in
          fol-plugin (rem3,lem2M)

----- gandalf can do lem1MS, but needs more than 3 s -----

lem1MS : (x : D) -> (y : D) -> (z:D) -> sup (meet x z) (meet y z) <= meet (sup x y) z
lem1MS = \x -> \y -> \z -> fol-plugin (ref, axS,axM,trans,lem1M,lem1S)



assocS : (x : D) -> (y : D) -> (z:D) -> Eq D (sup (sup x y) z) (sup x (sup y z))
assocS = \x -> \y -> \z -> fol-plugin (ref, axS)

comS : (x:D) -> (y:D) -> Eq D (sup x y) (sup y x)
comS = \x -> \y -> fol-plugin (ref, axS)

comM : (x:D) -> (y:D) -> Eq D (meet x y) (meet y x)
comM = \x -> \y -> fol-plugin (ref, axM)

---- an example which is not first-order -------

monotonic : (f:D -> D) -> Prop
monotonic = \ f -> (x:D) -> (y:D) -> x <= y ==> f x <= f y

meetP  : (f:D -> D) -> Prop
meetP = \ f -> (x:D) -> (y:D) -> Eq D (f (meet x y)) (meet (f x) (f y))

lemMon : (f:D -> D) -> meetP f -> monotonic f
lemMon = \ f -> \ h -> fol-plugin(lem2M,h)

testMon : (f:D -> D) -> monotonic f -> meetP f 
testMon = \ f -> \ h -> fol-plugin(lem2M,h,axM,ref,trans)


---- distributivity axiom

distr : (x:D) -> (y:D) -> (z:D) -> Eq D (meet x (sup y z)) (sup (meet x y) (meet x z))

lem3 : (x:D) -> (y:D) -> (z:D) -> (t:D) ->
        Eq D (meet (sup x y) (sup z t)) (sup (sup (meet x z) (meet y z)) (sup (meet x t) (meet y t)))
lem3 = 
 \ x -> \ y -> \ z -> \ t ->
  let
   rem : Eq D (meet (sup x y) (sup z t)) (sup (meet (sup x y) z) (meet (sup x y) t))
   rem = fol-plugin (distr)
   rem1 : Eq D (meet (sup x y) z) (sup (meet x z) (meet y z)) 
   rem1 = fol-plugin (distr,comS,comM)
   rem2 : Eq D (meet (sup x y) t) (sup (meet x t) (meet y t))
   rem2 = fol-plugin (distr,comS,comM)
  in 
   fol-plugin (rem,rem1,rem2)

lem4 : (x:D) -> (y:D) -> (z:D) -> (t:D) ->
        Eq D (meet (sup x y) (sup z t)) (sup (sup (meet x z) (meet x t)) (sup (meet y z) (meet y t)))
lem4 = fol-plugin (lem3,comS,assocS)

