-- Defines lists and some functions over them
-- Author :: Ulf Norell
-- Created:: 2004-12-06

-- # import Prelude
-- # import Nat
-- # import Ord

-- The list datatype ------------------------------------------------------
data List (A:Set) : Set where
  nil : List A
  (::) : A -> List A -> List A

infixr 10 ::
infixr 10 ++

foldr (A, B : Set) : (A -> B -> B) -> B -> List A -> B
foldr f e []      = e
foldr f e (x::xs)  = f x (foldr f e xs)

(++) (A:Set) : List A -> List A -> List A
(++) [] ys      = ys
(++) (x::xs) ys = x :: (xs ++ ys)

map (A, B : Set) : (A -> B) -> List A -> List B
map f []      = []
map f (x::xs)  = f x :: map f xs

length (A : Set) : List A -> Nat
length []     = 0
length (x::xs) = suc (length xs)

sum : List Nat -> Nat
sum []      = 0
sum (x::xs)  = x + sum xs

scanl (A, B : Set) : (A -> B -> A) -> A -> List B -> List A
scanl (*) e (x::xs)  = e :: scanl (*) (e * x) xs
scanl _ e []        = [e]

scanl1 (A : Set) : (A -> A -> A) -> List A -> List A
scanl1 op []      = []
scanl1 op (x::xs) = scanl op x xs

rev (A : Set) : List A -> List A -> List A
rev [] ys     = ys
rev (x::xs) ys = rev xs (x::ys)

reverse (A : Set) : List A -> List A
reverse xs = rev xs []


maximum (A : Set) : Ord A -> List A -> A
maximum ordA []      = epsilon
maximum ordA [x]     = x
maximum ordA (x::xs) = max ordA x (maximum ordA xs)

minimum (A : Set) : Ord A -> List A -> A
minimum ordA []      = epsilon
minimum ordA [x]     = x
minimum ordA (x::xs) = min ordA x (minimum ordA xs)

delete (A : Set) : Eq A -> A -> List A -> List A
delete equA y []      = []
delete equA y (x::xs) =
    if (equA.eq x y) xs (x :: delete equA y xs)



insert (A : Set) : Ord A -> A -> List A -> List A
insert ordA x [] = [x]
insert ordA x (y::ys) =
    if (ordA.leq x y)
       (x :: y :: ys)
       (y :: insert ordA x ys)

sortList (A : Set) : Ord A -> List A -> List A
sortList ordA [] = []
sortList ordA (x::xs) = insert ordA x (sortList ordA xs)
